#include "picture.h"

#include <QApplication>
#include <QDesktopWidget>
#include <QGraphicsScene>
#include <QPainter>
#include <QStyleOption>
#include <QDebug>
#include <QTimer>
#include <math.h>

#define ADJUST 10
#define PICTURE_SIZE 50
#define PICTURE_BORDER 5

Picture::Picture(QString path, float paramPx2mmScreen) : QGraphicsObject(), px2mmScreen(paramPx2mmScreen)
{
    setAcceptTouchEvents(true);

    img = QPixmap(path);

    if(img.width() > img.height()){
        x_width = PICTURE_SIZE/px2mmScreen;
        y_height = (x_width-2*PICTURE_BORDER) * img.height()/img.width() + 2*PICTURE_BORDER;
    }else{
        y_height = PICTURE_SIZE/px2mmScreen;
        x_width = (y_height-2*PICTURE_BORDER) * img.width()/img.height() + 2*PICTURE_BORDER;
    }
}

QRectF Picture::boundingRect() const
{
    return QRectF(-ADJUST, -ADJUST, x_width + 2*ADJUST, y_height + 2*ADJUST);
}

QPainterPath Picture::shape() const
{
    QPainterPath path;
    path.addRoundedRect(0,0, x_width, y_height, PICTURE_BORDER/px2mmScreen, PICTURE_BORDER/px2mmScreen);
    return path;
}

void Picture::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{
    painter->fillPath(shape(), QColor(0, 100, 150, 150));
    painter->drawPixmap(PICTURE_BORDER/px2mmScreen, PICTURE_BORDER/px2mmScreen,
                        x_width-2*PICTURE_BORDER/px2mmScreen, y_height-2*PICTURE_BORDER/px2mmScreen,
                        img);
    QPen pen;
    pen.setBrush(QColor(255, 255, 255, 153));
    pen.setWidth(1);
    painter->setPen(pen);
    painter->drawPath(shape());
}

bool Picture::sceneEvent(QEvent *event)
{
    QTouchEvent *touchEvent;

    switch (event->type()) {
    case QEvent::TouchBegin:
        //stackBefore(scene()->items().last());  //work send in background
        //stackBefore(scene()->items().first());  //TODO doesn't work
        foreach(QGraphicsItem *gi, scene()->items())  //mighty fix
            if(gi != this)
                gi->stackBefore(scene()->items().last());

    case QEvent::TouchUpdate:
    case QEvent::TouchEnd:
        touchEvent = static_cast<QTouchEvent *>(event);

        /*//Ignore touch outside of the shape
        foreach(QTouchEvent::TouchPoint touchPoint1, touchEvent->touchPoints())
            if(!contains(touchPoint1.lastPos()))
                return false;*/

        //Ignore touch outside of the shape
        {
        QList<QTouchEvent::TouchPoint> tmp;
        foreach(QTouchEvent::TouchPoint touchPoint1, touchEvent->touchPoints())
            if(contains(touchPoint1.lastPos()))
                tmp.append(touchPoint1);
        touchEvent->setTouchPoints(tmp);
        if(tmp.isEmpty())
            return false;
        }

        if(touchEvent->touchPoints().count() == 1) {
            const QTouchEvent::TouchPoint &touchPoint1 = touchEvent->touchPoints().first();
            moveBy(touchPoint1.scenePos().x() - touchPoint1.lastScenePos().x(),
                   touchPoint1.scenePos().y() - touchPoint1.lastScenePos().y());
        }else{
            const QTouchEvent::TouchPoint &touchPoint1 = touchEvent->touchPoints().first();
            const QTouchEvent::TouchPoint &touchPoint2 = touchEvent->touchPoints().last();
            QLineF line1(touchPoint1.lastScenePos(), touchPoint2.lastScenePos());
            QLineF line2(touchPoint1.scenePos(), touchPoint2.scenePos());

            float dAngle = line2.angleTo(line1);
            float dScale = line2.length() / line1.length();

            //Position de l'origine de l'image dans le repere 0
            QPointF origine0 = scenePos();
            //1) Determination du repere 1 associe au segment [lastA, lastB]
            //1.1) Calcul de la matrice de passage du repere 0 au repere 1
            QTransform m01(line1.dx(), line1.dy(),
                           -line1.dy(), line1.dx(),
                           touchPoint1.lastScenePos().x(), touchPoint1.lastScenePos().y());
            //1.2) Calcul de origine0 dans repere 1
            QTransform m10 = m01.inverted();
            QPointF origine1 = m10.map(origine0);

            //2) Calcul du repere 2 qui est celui associe au segment [currentA, currentB]
            QTransform m02(line2.dx(), line2.dy(),
                           -line2.dy(), line2.dx(),
                           touchPoint1.scenePos().x(), touchPoint1.scenePos().y());
            //La position de l'origine est la meme dans le repere 2 que celle dans le repere 1 au changement d'echelle pres.
            //m02 = m02.scale(dScale, dScale);  //pris en compte dans la construction de m02
            QPointF origine2 = origine1;
            QPointF origine0b = m02.map(origine2);
            QPointF dTrans = origine0b - origine0;

            moveBy(dTrans.x(), dTrans.y());
            setRotation(rotation()+dAngle);
            setScale(scale() * dScale);
        }
        break;

    default:
        return QGraphicsItem::sceneEvent(event);
    }
    return true;
}
