/*
 * Zombie - WidgetControl overlay manager for Qt
 *
 * Author: Jérémie Gilliot <jeremie.gilliot@inria.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Zombie.h"

#include <QPainter>
#include <QDebug>
#include <QGraphicsItem>
#include <QGraphicsView>
#include <QGraphicsWidget>
#include <QGraphicsProxyWidget>
#include <QPropertyAnimation>
#include <QScreen>
#include <QTimer>
#include <math.h>

#include "ghostview.h"
#include "qtuio.h"
#include "iostream"
#include "feedbackcreation.h"
#include "ripple.h"
#include "keyboardwidget.h"

#include "softtouchpad.h"


Zombie::Zombie(QWidget *parent) : QWidget(parent) {

    colors[0] = QColor(0,100, 150);  //bleu
    colors[1] = QColor(150,0, 0);  //rouge
    colors[2] = QColor(0,150, 0);  //vert
    colors[3] = QColor(200,200, 0);  //jaune
    colors[4] = QColor(200,65, 0);  //orange
    colors[5] = QColor(0,255, 0);  //vert fluo
    colors[6] = QColor(0,255, 200);  //bleu clair
    colors[7] = QColor(255,0, 255);  //rose
    colors[8] = QColor(150,100, 255);  //violet
    colors[9] = QColor(120,140, 115);  //gris
    colorIndex = 0;

    theParentWidget = qobject_cast<QWidget *>(parent);
    theParentView = qobject_cast<QGraphicsView *>(parent);
    if (theParentView)
        theParentScene = theParentView->scene();
    else
        theParentScene = qobject_cast<QGraphicsScene *>(parent);

    QGraphicsScene *scene = new QGraphicsScene(this);
    ghostView = new GhostView(scene, this, theParentWidget);
    setFixedSize(parent->size());
    ghostView->setFixedSize(parent->size());
    ghostView->setSceneRect(0, 0, parent->size().width(), parent->size().height());

    dissociateControlId = -1;
    dissociateCursorId = -1;
    ffbkCreate = new FeedbackCreation();
    ffbkCreate->setColor(colors[colorIndex]);
    ghostView->scene()->addItem(ffbkCreate);
    connect(ffbkCreate, SIGNAL(gestureCancelled()), this, SLOT(cancellingCreation()));
    connect(ffbkCreate, SIGNAL(gesturePerformed(QPointF,QPointF)), this, SLOT(createGhost(QPointF,QPointF)));

    this->px2mmScreen = 25.4 / QGuiApplication::primaryScreen()->physicalDotsPerInch();  //conversion dpi -> mm/px
    GhostView::setPx2mmScreen(px2mmScreen, 1);  //Used to compute circle size and give a scale for graphical elements
    ghostView->launchPrintScreenTimer();

    QApplication::instance()->installEventFilter(this);  //For ripple events
    theParentWidget->installEventFilter(this);  //For resize events

    setAttribute(Qt::WA_TransparentForMouseEvents);  //not reliable, because of mouse emulation

    qTUIO = new QTuio(ghostView);
    qTUIO->run();

    raise();
}

Zombie::~Zombie() {
    ghostView->scene()->removeItem(ffbkCreate);

    delete qTUIO;
    delete ghostView->scene();
    delete ghostView;

    QApplication::instance()->removeEventFilter(this);
    theParentWidget->removeEventFilter(this);
}

QList<QPointF> fingerFor5Click;

bool Zombie::eventFilter(QObject *obj, QEvent *event) {
    QWidget *target = qobject_cast<QWidget *>(obj);

    //Resize event
    if(event->type() == QEvent::Resize && target == theParentWidget) {
        QResizeEvent *rEvent = static_cast<QResizeEvent *>(event);
        setFixedSize(rEvent->size());
        ghostView->setFixedSize(rEvent->size());
        ghostView->setSceneRect(0, 0, rEvent->size().width(), rEvent->size().height());

    //Treat touch event in ghostView
    }else if((event->type() == QEvent::TouchBegin || event->type() == QEvent::TouchUpdate || event->type() == QEvent::TouchEnd) && target == ghostView->viewport()){
        QTouchEvent *tEvent = static_cast<QTouchEvent *>(event);
        foreach(QTouchEvent::TouchPoint touchPoint1, tEvent->touchPoints()){
            if(touchPoint1.state() == Qt::TouchPointPressed || touchPoint1.state() == Qt::TouchPointReleased)
                createRipple(target, touchPoint1);

            //Creation is reserved for real fingers on surface
            if(touchPoint1.id() >= 100)
                continue;



            //TODO implement retention of event while event can be a 5fingerClick
            //5clickFinger
            if(touchPoint1.state() == Qt::TouchPointPressed)
                fingerFor5Click.push_back(touchPoint1.pos());
            if(touchPoint1.state() == Qt::TouchPointMoved)
                if(QLineF(touchPoint1.pos(), touchPoint1.startPos()).length() > 15)
                    fingerFor5Click.removeAll(touchPoint1.startPos());
            if(touchPoint1.state() == Qt::TouchPointReleased){
                if(fingerFor5Click.size() == 4){
                    //Create new keyboard widget
                    QPointF widgetPos;
                    foreach(QPointF pt, fingerFor5Click){
                        widgetPos += pt;
                    }
                    widgetPos /= 4;
                    new KeyboardWidget(ghostView, widgetPos, px2mmScreen);
                }
                fingerFor5Click.clear();
            }



            if(touchPoint1.state() == Qt::TouchPointPressed){
                foreach(QTouchEvent::TouchPoint touchPoint2, tEvent->touchPoints()){
                    if(touchPoint2.id() != touchPoint1.id()
                    && touchPoint2.id() < 100
                    && QLineF(touchPoint1.pos(), touchPoint2.pos()).length() < 3*ghostView->circleSize
                    && QLineF(touchPoint2.pos(), touchPoint2.startPos()).length() < ghostView->circleSize){
                        dissociateControlId = touchPoint2.id();
                        dissociateCursorId = touchPoint1.id();
                        ffbkCreate->beginFingers(touchPoint1.pos(), touchPoint2.pos());
                        //return true;  //Retention of events while gesture could be a creation
                        break;
                    }
                }
            }else if(touchPoint1.state() == Qt::TouchPointMoved){
                if(touchPoint1.id() == dissociateControlId)
                    ffbkCreate->updateControl(touchPoint1.pos());
                else if(touchPoint1.id() == dissociateCursorId){
                    ffbkCreate->updateCursor(touchPoint1.pos());
                    return true;  //Retention of events while gesture could be a creation
                }
            }else if(touchPoint1.state() == Qt::TouchPointReleased){
                if(touchPoint1.id() == dissociateCursorId)
                    ffbkCreate->releaseCursor(touchPoint1.pos());
                else if(touchPoint1.id() == dissociateControlId)
                    ffbkCreate->releaseControl(touchPoint1.pos());
            }
        }
    }

    return QObject::eventFilter(obj, event);
}

void Zombie::cancellingCreation(){
    dissociateControlId = -1;
    dissociateCursorId = -1;
}

void Zombie::createRipple(QWidget *widget, const QTouchEvent::TouchPoint &touchPoint) {
    // if the Zombie has a parent
    if(theParentWidget->isAncestorOf(widget)) {
        QPoint newPoint = widget->mapTo(theParentWidget, QPoint(touchPoint.pos().x(), touchPoint.pos().y()));
        Ripple *ripple = new Ripple();
        ghostView->scene()->addItem(ripple);
        ripple->setPos(newPoint);
        ripple->setZValue(3);

        QPropertyAnimation *animation = new QPropertyAnimation(ripple, "radius");
        animation->setDuration(500);
        animation->setStartValue(ghostView->circleSize);
        animation->setEndValue(2*ghostView->circleSize);
        animation->setEasingCurve(QEasingCurve::Linear);
        connect(animation, SIGNAL(finished()), ripple, SLOT(deleteLater()));
        animation->start(QAbstractAnimation::DeleteWhenStopped);
    }
}

AbstractWidget *Zombie::createGhost(QPointF widgetPos, QPointF cursorPos)
{
    //Create a WallPad, dimension of 8*5cm
    QSize dim = QSize(80/px2mmScreen, 50/px2mmScreen);
    AbstractWidget *controlWidget = new SoftTouchpad(ghostView, colors[colorIndex++%10], widgetPos, cursorPos, dim);

    ffbkCreate->setColor(colors[colorIndex%10]);
    controlWidget->setParameters(px2mmScreen, dim*px2mmScreen);
    controlWidget->parentItem()->setRotation(0);
    return controlWidget;
}
