/*
 * Cursor - Touch cursor
 *
 * Author: Jérémie Gilliot <jeremie.gilliot@inria.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cursor.h"

#include <QApplication>
#include <QGraphicsScene>
#include <QGraphicsSceneMouseEvent>
#include <QPainter>
#include <QStyleOption>
#include <QTouchEvent>

#include "math.h"
#include "ghostview.h"
#include "QDebug"

#ifndef M_PI
#define M_PI       3.14159265358979323846
#define M_PI_2     1.57079632679489661923
#endif

#define ADJUST 10

//#define CIRCLE_SIZE (8/GhostView::px2mmScreen)  //30 px
#define CURSEL_SIZE (GhostView::scaleFactor*2/GhostView::px2mmScreen)  //5px
#define CLICK_HALO (GhostView::scaleFactor*10/GhostView::px2mmScreen)  //35px
#define CURSEL_HALO (GhostView::scaleFactor*4/GhostView::px2mmScreen)  //10px

static bool cursorsIds[100] = {false};  //false means id is free


Cursor::Cursor(GhostView *g) : QGraphicsObject(), ghostview(g)
{
    cursorNb = takeNumber();

    leftClick = false;
    oldLeftClick = false;

    firstId = -1;
    eventCenterId = -1;

    setAcceptTouchEvents(true);
    //setOpacity(0.3);
}

Cursor::~Cursor(){
    setLeftClick(false);

    foreach(int id, curselList.keys())
        releaseNumber(id);
    releaseNumber(cursorNb);
}

int Cursor::getCursorNb(){
    return cursorNb;
}

bool Cursor::isIdPartOfCursor(int id){
    return id == cursorNb || curselList.contains(id);
}

int Cursor::takeNumber(){
    int number = -1;
    for(int i=0; i<100; i++){
        if(!cursorsIds[i]){
            number = i+200;
            cursorsIds[i] = true;
            break;
        }
        if(i == 99){
            qErrnoWarning("More than 100 cursors and cursels");
            return -1;
            //exit(EXIT_FAILURE);
        }
    }
    return number;
}

void Cursor::releaseNumber(int id){
    cursorsIds[id-200] = false;
}


QRectF Cursor::boundingRect() const
{
    QRectF br(-GhostView::circleSize - CURSEL_SIZE/2 - ADJUST, -GhostView::circleSize - CURSEL_SIZE/2 - ADJUST,
              2*GhostView::circleSize + CURSEL_SIZE + 2*ADJUST, 2*GhostView::circleSize + CURSEL_SIZE + 2*ADJUST);

    //Add cursels
    foreach(int id, curselList.keys()){
        QPointF cursel = curselList[id];
        if(id == firstId)
             continue;

        QRectF curselBR(- CURSEL_SIZE/2 - ADJUST, - CURSEL_SIZE/2 - ADJUST,
                        CURSEL_SIZE + 2*ADJUST, CURSEL_SIZE + 2*ADJUST);
        curselBR.translate(cursel-firstTouch);
        br = br.united(curselBR);
    }

    //Add removed cursels
    foreach(QPointF cursel, oldCurselList){
        QRectF curselBR(- CURSEL_SIZE/2 - ADJUST, - CURSEL_SIZE/2 - ADJUST,
                        CURSEL_SIZE + 2*ADJUST, CURSEL_SIZE + 2*ADJUST);
        curselBR.translate(cursel-firstTouch);
        br = br.united(curselBR);
    }

    //return br.adjusted(-ADJUST, -ADJUST, ADJUST, ADJUST);
    br.adjust(-ADJUST, -ADJUST, ADJUST, ADJUST);
    return br.united(br.translated(oldPos-pos()));
}

QPainterPath Cursor::shape() const
{
    QPainterPath path;
    path.addEllipse(-GhostView::circleSize, -GhostView::circleSize, GhostView::circleSize * 2, GhostView::circleSize * 2);
    return path;
}

void Cursor::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{
    oldCurselList.clear();

    QPen pen;
    pen.setBrush(Qt::white);
    pen.setWidth(1);
    painter->setPen(pen);

    //Paint click halo
    if(leftClick){
        QPainterPath blob;
        blob.setFillRule(Qt::WindingFill);
        blob.addEllipse(-CLICK_HALO, -CLICK_HALO, 2*CLICK_HALO, 2*CLICK_HALO);
        foreach(int id, curselList.keys()){
            QPointF cursel = curselList[id];
            if(id == firstId)
                 continue;
            blob.addEllipse(cursel-firstTouch, CURSEL_HALO, CURSEL_HALO);
        }
        painter->drawPath(blob);
        QColor fade = color;
        fade.setAlpha(100);
        painter->fillPath(blob, fade);
    }

    //Paint cursor
    /*QPainterPath cursorCircle;
    cursorCircle.addEllipse(-GhostView::circleSize, -GhostView::circleSize, GhostView::circleSize * 2, GhostView::circleSize * 2);
    painter->fillPath(cursorCircle, color);
    painter->drawEllipse(-GhostView::circleSize, -GhostView::circleSize, GhostView::circleSize * 2, GhostView::circleSize * 2) ;
    painter->drawLine(-CURSEL_SIZE, 0, CURSEL_SIZE, 0);
    painter->drawLine(0, -CURSEL_SIZE, 0, CURSEL_SIZE);*/

    QPainterPath cursorArrow;
    QVector<QPointF> pts;
    pts.append(QPointF(0, 0));
    pts.append(QPointF(0, GhostView::circleSize));
    pts.append(QPointF(GhostView::circleSize/4.0, 1.73*GhostView::circleSize/2.0));
    pts.append(QPointF(GhostView::circleSize/2.0, 0.8*1.73*GhostView::circleSize));
    pts.append(QPointF(GhostView::circleSize/2.0 + GhostView::circleSize/8.0, 0.8*(1.73*GhostView::circleSize - GhostView::circleSize/20.0)));
    pts.append(QPointF(GhostView::circleSize/4.0 + GhostView::circleSize/8.0, 1.73*GhostView::circleSize/2.0 - GhostView::circleSize/20.0));
    pts.append(QPointF(1.41*GhostView::circleSize/2.0, 1.41*GhostView::circleSize/2.0));
    pts.append(QPointF(0, 0));
    QPolygonF poly(pts);
    cursorArrow.addPolygon(poly);
    painter->fillPath(cursorArrow, color);
    painter->drawPath(cursorArrow);

#ifndef EXPE_PROPRIO

#else
    // EXPE Cross lines for experiment proprioception
    painter->drawLine(-GhostView::circleSize, 0, GhostView::circleSize, 0);
    painter->drawLine(0, -GhostView::circleSize, 0, GhostView::circleSize);
#endif

    //Paint cursels
    foreach(int id, curselList.keys()){
        QPointF cursel = curselList[id];
        if(id == firstId)
             continue;
        QPainterPath curselCircles;
        curselCircles.addEllipse(cursel-firstTouch, CURSEL_SIZE, CURSEL_SIZE);
        painter->fillPath(curselCircles, color);
        painter->drawEllipse(cursel-firstTouch, CURSEL_SIZE, CURSEL_SIZE);

        QLineF l(QPointF(0,0) , cursel-firstTouch);
        if(l.length() > GhostView::circleSize){
            float ratio = GhostView::circleSize / l.length();
            painter->drawLine((cursel-firstTouch)*ratio , cursel-firstTouch);
        }
    }

    oldPos = pos();
}

bool Cursor::sceneEvent(QEvent *event)
{
    QTouchEvent *touchEvent;

    switch (event->type()) {
    case QEvent::TouchBegin:
    case QEvent::TouchUpdate:
    case QEvent::TouchEnd:
        touchEvent = static_cast<QTouchEvent *>(event);
        if (touchEvent->touchPoints().size() == 1) {
            const QTouchEvent::TouchPoint &touchPoint1 = touchEvent->touchPoints().first();
            if(isIdPartOfCursor(touchPoint1.id())){  //if event is generated by itself, return false to pass to underlying item. item's existence is ensured in ghostview dispatching
                return false;  //reject event
            }
            if(touchPoint1.state() == Qt::TouchPointPressed)
                receiveTouchBegin();
            else if(touchPoint1.state() == Qt::TouchPointReleased)
                receiveTouchEnd();
            moveBy(touchPoint1.scenePos().x() - touchPoint1.lastScenePos().x(),
                   touchPoint1.scenePos().y() - touchPoint1.lastScenePos().y());
        }
        sendUpdateEvent();
        generateEvent();
        return true;

    default:
        return QGraphicsItem::sceneEvent(event);
    }
}

void Cursor::setLeftClick(bool c){
    struct touchWidget tw;

    leftClick = c;
    if(leftClick && !oldLeftClick){
        //Cursor
        if(curselList.size()==1){
            centerClick = true;
            eventCenterId = cursorNb;
            QTouchEvent::TouchPoint newTouchPoint(eventCenterId);
            newTouchPoint.setState(Qt::TouchPointPressed);
            newTouchPoint.setScreenPos(pos() + ghostview->mapToGlobal(QPoint()) - ghostview->parentWidget()->pos());
            newTouchPoint.setNormalizedPos(QPointF(newTouchPoint.screenPos().x()/(float)ghostview->height(), newTouchPoint.screenPos().y()/(float)ghostview->width()));

            tw.tPoint = newTouchPoint; tw.widget = ghostview->viewport();
            touchPointList.insert(eventCenterId, tw);
            updateTouch(&newTouchPoint);
        }

        //Cursels
        foreach(int id, curselList.keys()){
            if(id == firstId)
                continue;
            QTouchEvent::TouchPoint newTouchPoint(id);
            newTouchPoint.setState(Qt::TouchPointPressed);
            newTouchPoint.setScreenPos(mapToScene(curselList[id]-firstTouch) + ghostview->mapToGlobal(QPoint()) - ghostview->parentWidget()->pos());
            newTouchPoint.setNormalizedPos(QPointF(newTouchPoint.screenPos().x()/(float)ghostview->height(), newTouchPoint.screenPos().y()/(float)ghostview->width()));

            tw.tPoint = newTouchPoint; tw.widget = ghostview->viewport();
            touchPointList.insert(id, tw);
            updateTouch(&newTouchPoint);
        }

        generateEvent();

    }else if(!leftClick && oldLeftClick){
        //Cursor
        if(centerClick){
            centerClick = false;
            QTouchEvent::TouchPoint newTouchPoint(eventCenterId);
            newTouchPoint.setState(Qt::TouchPointReleased);
            newTouchPoint.setScreenPos(pos() + ghostview->mapToGlobal(QPoint()) - ghostview->parentWidget()->pos());
            newTouchPoint.setNormalizedPos(QPointF(newTouchPoint.screenPos().x()/(float)ghostview->height(), newTouchPoint.screenPos().y()/(float)ghostview->width()));

            tw = touchPointList[eventCenterId];
            tw.tPoint.setState(newTouchPoint.state());
            updateTouch(&newTouchPoint);
            touchPointList.insert(eventCenterId, tw);
            if((eventCenterId != firstId) && (eventCenterId != cursorNb))
                releaseNumber(eventCenterId);
            eventCenterId = -1;
        }

        //Cursels
        foreach(int id, curselList.keys()){
            if(id == firstId)
                continue;
            QTouchEvent::TouchPoint newTouchPoint(id);
            newTouchPoint.setState(Qt::TouchPointReleased);
            newTouchPoint.setScreenPos(mapToScene(curselList[id]-firstTouch) + ghostview->mapToGlobal(QPoint()) - ghostview->parentWidget()->pos());
            newTouchPoint.setNormalizedPos(QPointF(newTouchPoint.screenPos().x()/(float)ghostview->height(), newTouchPoint.screenPos().y()/(float)ghostview->width()));

            tw = touchPointList[id];
            tw.tPoint.setState(newTouchPoint.state());
            updateTouch(&newTouchPoint);
            touchPointList.insert(id, tw);
        }

        generateEvent();
    }
    oldLeftClick = leftClick;
}

void Cursor::sendUpdateEvent(){
    struct touchWidget tw;

    if(leftClick){
        //Cursor
        if(centerClick){
            QTouchEvent::TouchPoint newTouchPoint(eventCenterId);
            tw = touchPointList[eventCenterId];
            newTouchPoint.setState(Qt::TouchPointMoved);
            newTouchPoint.setScreenPos(pos() + ghostview->mapToGlobal(QPoint()) - ghostview->parentWidget()->pos());
            newTouchPoint.setNormalizedPos(QPointF(newTouchPoint.screenPos().x()/(float)ghostview->height(), newTouchPoint.screenPos().y()/(float)ghostview->width()));

            tw.tPoint.setState(newTouchPoint.state());
            touchPointList.insert(eventCenterId, tw);
            updateTouch(&newTouchPoint);
        }

        //Cursels
        foreach(int id, curselList.keys()){
            if(id == firstId)
                continue;
            QTouchEvent::TouchPoint newTouchPoint(id);
            tw = touchPointList[id];
            newTouchPoint.setState(Qt::TouchPointMoved);
            newTouchPoint.setScreenPos(mapToScene(curselList[id]-firstTouch) + ghostview->mapToGlobal(QPoint()) - ghostview->parentWidget()->pos());
            newTouchPoint.setNormalizedPos(QPointF(newTouchPoint.screenPos().x()/(float)ghostview->height(), newTouchPoint.screenPos().y()/(float)ghostview->width()));

            tw.tPoint.setState(newTouchPoint.state());
            touchPointList.insert(id, tw);
            updateTouch(&newTouchPoint);
        }

        //generateEvent();
    }
}

void Cursor::generateEvent(){
    bool touchBegin = true;
    bool touchEnd = true;
    bool sendEvent = false;
    Qt::TouchPointStates tps;
    QList<QTouchEvent::TouchPoint> tPts;

    //Add touches already exist in ghostview
    foreach(QTouchEvent::TouchPoint pt, ghostview->getTouchpointList()){
        if(pt.state() != Qt::TouchPointReleased && !touchPointList.keys().contains(pt.id())){
            touchBegin = false;
            touchEnd = false;
            pt.setState(Qt::TouchPointStationary);
            tps |= pt.state();
            tPts.append(pt);
        }
    }

    foreach(int id, touchPointList.keys()){
        struct touchWidget tw = touchPointList[id];
        if(tw.tPoint.state() != Qt::TouchPointPressed)
            touchBegin = false;
        if(tw.tPoint.state() != Qt::TouchPointReleased)
            touchEnd = false;
        if(tw.tPoint.state() != Qt::TouchPointStationary)
            sendEvent = true;
        tps |= tw.tPoint.state();
        tPts.append(tw.tPoint);

        if(tw.tPoint.state() == Qt::TouchPointReleased){
            touchPointList.remove(id);
        }else{
            tw.tPoint.setState(Qt::TouchPointMoved);
            touchPointList.insert(id, tw);
        }
    }

    if(sendEvent){
        QEvent::Type eventType;
        if(touchBegin)
            eventType = QEvent::TouchBegin;
        else if(touchEnd)
            eventType = QEvent::TouchEnd;
        else
            eventType = QEvent::TouchUpdate;
        QEvent *touchEvent = new QTouchEvent(eventType, new QTouchDevice() /*QTouchDevice::TouchScreen*/, Qt::NoModifier, tps, tPts);
        //qDebug("Cursor %d send %d nb=%d to ghostview", cursorNb, eventType, tPts.size());
        qApp->sendEvent(ghostview->viewport(), touchEvent);
    }
    update();
}

void Cursor::updateTouch(QTouchEvent::TouchPoint *newTouchPoint){
    if(!touchPointList.contains(newTouchPoint->id()))
        return;
    struct touchWidget tw = touchPointList[newTouchPoint->id()];

    // --- Normalized coordinates ---
    tw.tPoint.setLastNormalizedPos(tw.tPoint.normalizedPos());
    tw.tPoint.setNormalizedPos(newTouchPoint->normalizedPos());
    // --- Screen coordinates ---
    tw.tPoint.setLastScreenPos(tw.tPoint.screenPos());
    tw.tPoint.setScreenPos(newTouchPoint->screenPos());
    // --- Widget coordinates ---
    tw.tPoint.setLastPos(tw.tPoint.pos());
    tw.tPoint.setPos(newTouchPoint->screenPos() + tw.widget->mapFromGlobal(QPoint()));
    //tw.tPoint.setPos(tw.widget->mapFromGlobal(QPoint(newTouchPoint->screenPos().x(), newTouchPoint->screenPos().y())));
    //// --- Widget coordinates ---
    //tw.tPoint.setLastScenePos(tw.tPoint.scenePos());
    //tw.tPoint.setScenePos(tw.widget->mapFromGlobal(QPoint(touchPoint1.screenPos().x(), touchPoint1.screenPos().y())));

    // --- Start coordinates ---
    if(newTouchPoint->state() == Qt::TouchPointPressed){
        tw.tPoint.setStartNormalizedPos(tw.tPoint.normalizedPos());
        tw.tPoint.setStartPos(tw.tPoint.pos());
        tw.tPoint.setStartScreenPos(tw.tPoint.screenPos());
        tw.tPoint.setStartScenePos(tw.tPoint.scenePos());

        tw.tPoint.setLastNormalizedPos(tw.tPoint.normalizedPos());
        tw.tPoint.setLastPos(tw.tPoint.pos());
        tw.tPoint.setLastScreenPos(tw.tPoint.screenPos());
        tw.tPoint.setLastScenePos(tw.tPoint.scenePos());
    }
    touchPointList.insert(newTouchPoint->id(), tw);
}

int Cursor::addCursel(QPointF fPos){
    //qDebug("    addCursel(%f,%f)", fPos.x(),fPos.y());
    fPos = GhostView::scaleFactor * fPos;
    int fingerId = takeNumber();
    if(fingerId == -1)
        return -1;

    if(firstId == -1){
        firstId = fingerId;
        if(curselList.isEmpty()){  //improve the stability of cursor and allow to clutch with primal finger
            foreach(QPointF cursel, curselList.values())
                oldCurselList.append(cursel+fPos-firstTouch);
            firstTouch = fPos;
        }
    }
    curselList.insert(fingerId, fPos);

    if(leftClick){
        if(fingerId == firstId){
            sendUpdateEvent();
        }else{
            QTouchEvent::TouchPoint newTouchPoint(fingerId);
            newTouchPoint.setState(Qt::TouchPointPressed);
            newTouchPoint.setScreenPos(mapToScene(curselList[fingerId]-firstTouch) + ghostview->mapToGlobal(QPoint()) - ghostview->parentWidget()->pos());
            newTouchPoint.setNormalizedPos(QPointF(newTouchPoint.screenPos().x()/(float)ghostview->height(), newTouchPoint.screenPos().y()/(float)ghostview->width()));

            struct touchWidget tw;
            tw.tPoint = newTouchPoint; tw.widget = ghostview->viewport();
            touchPointList.insert(fingerId, tw);
            updateTouch(&newTouchPoint);
            generateEvent();
        }
    }

    return fingerId;
}

void Cursor::moveCursel(int id, QPointF fPos){
    //qDebug("    moveCursel(%f,%f)", fPos.x(),fPos.y());
    fPos = GhostView::scaleFactor * fPos;
    //Keep cursor in screen
    if(pos().x() < 0) setPos(0, pos().y());
    if(pos().x() > ghostview->width()) setPos(ghostview->width(), pos().y());
    if(pos().y() < 0) setPos(pos().x(), 0);
    if(pos().y() > ghostview->height()) setPos(pos().x(), ghostview->height());
    //moveBy(fPos.x()/curselList.size(), fPos.y()/curselList.size());
    if(id == firstId)
        firstTouch += fPos;
    curselList.insert(id, curselList[id] + fPos);

    if(leftClick){
        if(id == firstId){
            sendUpdateEvent();
        }else{
            QTouchEvent::TouchPoint newTouchPoint(id);
            struct touchWidget tw = touchPointList[id];
            /*if(tw.tPoint.screenPos() == ghostview->mapToGlobal(mapToScene(curselList[id]-firstTouch).toPoint()))
                newTouchPoint.setState(Qt::TouchPointStationary);
            else*/
                newTouchPoint.setState(Qt::TouchPointMoved);
            newTouchPoint.setScreenPos(mapToScene(curselList[id]-firstTouch) + ghostview->mapToGlobal(QPoint()) - ghostview->parentWidget()->pos());
            newTouchPoint.setNormalizedPos(QPointF(newTouchPoint.screenPos().x()/(float)ghostview->height(), newTouchPoint.screenPos().y()/(float)ghostview->width()));

            tw.tPoint.setState(newTouchPoint.state());
            touchPointList.insert(id, tw);
            updateTouch(&newTouchPoint);
            generateEvent();
        }
    }
}

void Cursor::removeCursel(int id, QPointF fPos){
    //qDebug("    removeCursel(%f,%f)", fPos.x(),fPos.y());
    fPos = GhostView::scaleFactor * fPos;
    if(id == firstId)
        firstTouch += fPos;
    curselList.insert(id, curselList[id] + fPos);

    if(leftClick){
        if(id == firstId){
            sendUpdateEvent();
        }else{
            QTouchEvent::TouchPoint newTouchPoint(id);
            newTouchPoint.setState(Qt::TouchPointReleased);
            newTouchPoint.setScreenPos(mapToScene(curselList[id]-firstTouch) + ghostview->mapToGlobal(QPoint()) - ghostview->parentWidget()->pos());
            newTouchPoint.setNormalizedPos(QPointF(newTouchPoint.screenPos().x()/(float)ghostview->height(), newTouchPoint.screenPos().y()/(float)ghostview->width()));

            struct touchWidget tw;
            tw = touchPointList[id];
            tw.tPoint.setState(newTouchPoint.state());
            updateTouch(&newTouchPoint);
            touchPointList.insert(id, tw);
            generateEvent();
        }
    }

    oldCurselList.append(curselList[id]);
    releaseNumber(id);
    curselList.remove(id);
    if(id == firstId)
        firstId = -1;
}
