/*
 * GhostView - WidgetControl overlay for Qt
 *
 * Author: Jérémie Gilliot <jeremie.gilliot@inria.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ghostview.h"

#include <QApplication>
#include <QDesktopWidget>
#include <QScreen>
#include <QGraphicsItem>
#include <QGraphicsWidget>
#include <QTimer>
#include <QDebug>
#include <math.h>

#include "cursor.h"


GhostView::GhostView(QGraphicsScene *scene, QWidget *parent, QWidget *uA) : QGraphicsView(scene, parent){
    underApp = qobject_cast<QWidget *>(uA);
    underAppView = qobject_cast<QGraphicsView *>(uA);
    if (underAppView)
        underAppScene = underAppView->scene();
    else
        underAppScene = qobject_cast<QGraphicsScene *>(uA);

    setStyleSheet("background: transparent");
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setAttribute(Qt::WA_TransparentForMouseEvents);
    viewport()->setAttribute(Qt::WA_AcceptTouchEvents);
    viewport()->setAttribute(Qt::WA_TransparentForMouseEvents);
    setRenderHints(QPainter::Antialiasing);
    setViewportUpdateMode(QGraphicsView::BoundingRectViewportUpdate);  //TODO regler les bounding box pour l'update de la bonne zone

    timerPrintScreen = new QTimer(this);
    connect(timerPrintScreen, SIGNAL(timeout()), this, SLOT(updatePrintScreen()));
    connect(this, SIGNAL(destroyed()), timerPrintScreen, SLOT(deleteLater()));
    updatePrintScreen();
}

float GhostView::px2mmScreen;
float GhostView::scaleFactor;
float GhostView::circleSize;

void GhostView::setPx2mmScreen(float px, float sf){
    GhostView::px2mmScreen = px;
    GhostView::scaleFactor = sf;
    GhostView::circleSize = (GhostView::scaleFactor*8/GhostView::px2mmScreen);
}

void GhostView::launchPrintScreenTimer(){
    timerPrintScreen->start(200);  //grabWindow cause a light freeze of screen : unsuitable !!
}

void GhostView::stopPrintScreenTimer(){
    timerPrintScreen->stop();
}

void GhostView::updatePrintScreen(){
    QPoint origin = mapToGlobal(QPoint());
    radar = QGuiApplication::primaryScreen()->grabWindow(QApplication::desktop()->winId(), origin.x(), origin.y(), width(), height());
}

QPixmap GhostView::getPrintScreen(){
    return radar;
}

bool GhostView::viewportEvent(QEvent *event){
    QWidget *newTarget;
    QTouchEvent *tEvent;
    QList<QTouchEvent::TouchPoint> listPtsOnGhostItem;
    Qt::TouchPointStates stateGhost;
    QTouchEvent *eventGhost;

    switch(event->type()){
    case QEvent::TouchBegin:
    case QEvent::TouchUpdate:
    case QEvent::TouchEnd:
        tEvent = static_cast<QTouchEvent *>(event);
        //qDebug("START %d touches", tEvent->touchPoints().size());
        foreach(QTouchEvent::TouchPoint touchPoint1, tEvent->touchPoints()){
            QGraphicsItem *select;
            QList<QGraphicsItem *> selectList;
            QTouchEvent::TouchPoint newTouchPoint(touchPoint1);
            newTouchPoint.setId(touchPoint1.id()+500);

            QList<int> widgetTouches;
            struct touchWidget tw;
            //qDebug("touchpointState=%d", touchPoint1.state());
            switch(touchPoint1.state()){
            case Qt::TouchPointPressed:
                //qDebug("Ghost recv BEG , nb de points=%d, id=%d", tEvent->touchPoints().size(), touchPoint1.id());
                listPtsOnGhostView.insert(touchPoint1.id(), touchPoint1);

                selectList = items(touchPoint1.pos().toPoint());
                //test if touch is on ghost item
                //limit ghostview interaction for real finger (no cursor and cursels)
                if(touchPoint1.id()<200 && !selectList.isEmpty()){  //if touch grab item
                    select = selectList.first();
                    Cursor* c = dynamic_cast<Cursor*>(select);
                    /*if(c)
                        qDebug("TEST if cursel %d appartient au curseur %d = %d", touchPoint1.id(), c->getCursorNb(), c->isIdPartOfCursor(touchPoint1.id()));*/
                    if( c == NULL  //if select is NOT a cursor
                     || c->isIdPartOfCursor(touchPoint1.id()) == false  //or if select is NOT the cursor which create the event
                     || selectList.size() > 1 ){  //or if select IS the cursor which create the event BUT underlying ghostView item EXISTS
                        //qDebug("  GhostItem found");
                        listPtsOnGhostItem.append(touchPoint1);
                        stateGhost |= touchPoint1.state();
                        break;
                    }
                }

                //else pass to the underneath application
                parentWidget()->setVisible(false);
                newTarget = underApp->childAt(touchPoint1.pos().toPoint());
                if(newTarget == 0) newTarget = underApp;
                parentWidget()->setVisible(true);

                tw.tPoint = newTouchPoint; tw.widget = newTarget;
                touchPointList.insert(newTouchPoint.id(), tw);
                if(widgets.contains(newTarget))
                    widgetTouches = widgets[newTarget];
                widgetTouches.append(newTouchPoint.id());
                widgets.insert(newTarget, widgetTouches);
                updateTouch(&touchPoint1);
                if(touchPoint1.id() == tEvent->touchPoints().last().id() /*isPrimary()*/ /*&& !tw.widget->testAttribute(Qt::WA_AcceptTouchEvents)*/){
                    struct touchWidget tmp = touchPointList[newTouchPoint.id()];
                    //qDebug("primary press pos(%f,%f) globalPos(%f,%f)", tmp.tPoint.pos().x(),tmp.tPoint.pos().y(), tmp.tPoint.screenPos().x(),tmp.tPoint.screenPos().y());
                    QMouseEvent *mouseEmul = new QMouseEvent(QEvent::MouseButtonPress, QPoint(tmp.tPoint.pos().x(),tmp.tPoint.pos().y()), QPoint(tmp.tPoint.screenPos().x(),tmp.tPoint.screenPos().y()), Qt::LeftButton, Qt::LeftButton, Qt::NoModifier);
                    qApp->postEvent(tmp.widget, mouseEmul);
                }
                //qDebug("  underneathTargetPosition (%d, %d)", newTarget->pos().x(), newTarget->pos().y());
                break;

            case Qt::TouchPointMoved:
            case Qt::TouchPointStationary:
                //qDebug("Ghost recv UPD , nb de points=%d, id=%d", tEvent->touchPoints().size(), touchPoint1.id());
                listPtsOnGhostView.insert(touchPoint1.id(), touchPoint1);
                if(touchPointList.contains(touchPoint1.id()+500)){
                    //pass to the underneath application
                    tw = touchPointList[newTouchPoint.id()];
                    tw.tPoint.setState(touchPoint1.state());
                    touchPointList.insert(newTouchPoint.id(), tw);
                    updateTouch(&touchPoint1);
                    if(touchPoint1.id() == tEvent->touchPoints().last().id() /*isPrimary()*/ /*&& !tw.widget->testAttribute(Qt::WA_AcceptTouchEvents)*/){
                        struct touchWidget tmp = touchPointList[newTouchPoint.id()];
                        //qDebug("primary drag pos(%f,%f) globalPos(%f,%f)", tmp.tPoint.pos().x(),tmp.tPoint.pos().y(), tmp.tPoint.screenPos().x(),tmp.tPoint.screenPos().y());
                        QMouseEvent *mouseEmul = new QMouseEvent(QEvent::MouseMove, QPoint(tmp.tPoint.pos().x(),tmp.tPoint.pos().y()), QPoint(tmp.tPoint.screenPos().x(),tmp.tPoint.screenPos().y()), Qt::NoButton, Qt::LeftButton, Qt::NoModifier);
                        qApp->postEvent(tmp.widget, mouseEmul);
                    }
                }else{
                    listPtsOnGhostItem.append(touchPoint1);
                    stateGhost |= touchPoint1.state();
                }
                break;

            case Qt::TouchPointReleased:
                //qDebug("Ghost recv END , nb de points=%d, id=%d", tEvent->touchPoints().size(), touchPoint1.id());
                listPtsOnGhostView.remove(touchPoint1.id());
                if(touchPointList.contains(touchPoint1.id()+500)){
                    //pass to the underneath application
                    tw = touchPointList[newTouchPoint.id()];
                    tw.tPoint.setState(touchPoint1.state());
                    updateTouch(&touchPoint1);
                    touchPointList.insert(newTouchPoint.id(), tw);
                    if(touchPoint1.id() == tEvent->touchPoints().last().id() /*isPrimary()*/ /*&& !tw.widget->testAttribute(Qt::WA_AcceptTouchEvents)*/){
                        struct touchWidget tmp = touchPointList[newTouchPoint.id()];
                        //qDebug("primary release pos(%f,%f) globalPos(%f,%f)", tmp.tPoint.pos().x(),tmp.tPoint.pos().y(), tmp.tPoint.screenPos().x(),tmp.tPoint.screenPos().y());
                        QMouseEvent *mouseEmul = new QMouseEvent(QEvent::MouseButtonRelease, QPoint(tmp.tPoint.pos().x(),tmp.tPoint.pos().y()), QPoint(tmp.tPoint.screenPos().x(),tmp.tPoint.screenPos().y()), Qt::LeftButton, Qt::NoButton, Qt::NoModifier);
                        qApp->postEvent(tmp.widget, mouseEmul);
                    }
                }else{
                    listPtsOnGhostItem.append(touchPoint1);
                    stateGhost |= touchPoint1.state();
                }
                break;

            default:
                break;
            }
        }

        foreach(QWidget *currWidget, widgets.keys()){
            bool touchBegin = true;
            bool touchEnd = true;
            bool sendEvent = false;
            Qt::TouchPointStates tps;
            QList<QTouchEvent::TouchPoint> tPts;


            foreach(int id, widgets[currWidget]){
                struct touchWidget tw = touchPointList[id];
                if(tw.tPoint.state() != Qt::TouchPointPressed)
                    touchBegin = false;
                if(tw.tPoint.state() != Qt::TouchPointReleased)
                    touchEnd = false;
                if(tw.tPoint.state() != Qt::TouchPointStationary)
                    sendEvent = true;
                tps |= tw.tPoint.state();
                tPts.append(tw.tPoint);

                if(tw.tPoint.state() == Qt::TouchPointReleased){
                    touchPointList.remove(id);
                    widgets[currWidget].removeAll(id);
                }else{
                    tw.tPoint.setState(Qt::TouchPointStationary);
                    touchPointList.insert(id, tw);
                }
            }

            if(sendEvent){
                QEvent::Type eventType;
                if(touchBegin){
                    eventType = QEvent::TouchBegin;
                }else if(touchEnd){
                    eventType = QEvent::TouchEnd;
                }else{
                    eventType = QEvent::TouchUpdate;
                }
                QEvent *touchEvent = new QTouchEvent(eventType, new QTouchDevice() /*QTouchDevice::TouchScreen*/, Qt::NoModifier, tps, tPts);

                /*qDebug("Ghostview send %d nb=%d to underApp", eventType, tPts.size());
                foreach(QTouchEvent::TouchPoint t, tPts)
                    qDebug("\tpos(%3.2f, %3.2f) state=%d", t.pos().x(), t.pos().y(), t.state());*/

                if(underAppView && underAppView->viewport())
                  qApp->postEvent(underAppView->viewport(), touchEvent);
                else if(underAppScene)
                  qApp->postEvent(underAppScene, touchEvent);
                else
                  qApp->postEvent(currWidget, touchEvent);
            }
        }

        //if a widget has an empty list it can be dropped
        foreach(QWidget *currWidget, widgets.keys())
            if(widgets[currWidget].isEmpty())
                widgets.remove(currWidget);


        //Treat events concerning ghost items
        QEvent::Type eventType;
        if(stateGhost && !Qt::TouchPointPressed == 0)
            eventType = QEvent::TouchBegin;
        else if(stateGhost && !Qt::TouchPointReleased == 0)
            eventType = QEvent::TouchEnd;
        else
            eventType = QEvent::TouchUpdate;
        eventGhost = new QTouchEvent(eventType, tEvent->device(), tEvent->modifiers(), stateGhost, listPtsOnGhostItem);
        QGraphicsView::viewportEvent(eventGhost);

        //Accept all touch events received
        return true;

    default:
        break;
    }
    return QGraphicsView::viewportEvent(event);
}

void GhostView::updateTouch(QTouchEvent::TouchPoint *newTouchPoint){
    if(!touchPointList.contains(newTouchPoint->id()+500))
        return;
    struct touchWidget tw = touchPointList[newTouchPoint->id()+500];

    // --- Normalized coordinates ---
    tw.tPoint.setLastNormalizedPos(tw.tPoint.normalizedPos());
    tw.tPoint.setNormalizedPos(newTouchPoint->normalizedPos());
    // --- Screen coordinates ---
    tw.tPoint.setLastScreenPos(tw.tPoint.screenPos());
    tw.tPoint.setScreenPos(newTouchPoint->screenPos());
    // --- Widget coordinates ---
    tw.tPoint.setLastPos(tw.tPoint.pos());
    tw.tPoint.setPos(newTouchPoint->screenPos() + tw.widget->mapFromGlobal(QPoint()));
    /*// --- Widget coordinates ---
    tw.tPoint.setLastScenePos(tw.tPoint.scenePos());
    tw.tPoint.setScenePos(tw.widget->mapFromGlobal(QPoint(touchPoint1.screenPos().x(), touchPoint1.screenPos().y())));*/

    // --- Start coordinates ---
    if(newTouchPoint->state() == Qt::TouchPointPressed){
        tw.tPoint.setStartNormalizedPos(tw.tPoint.normalizedPos());
        tw.tPoint.setStartPos(tw.tPoint.pos());
        tw.tPoint.setStartScreenPos(tw.tPoint.screenPos());
        tw.tPoint.setStartScenePos(tw.tPoint.scenePos());

        tw.tPoint.setLastNormalizedPos(tw.tPoint.normalizedPos());
        tw.tPoint.setLastPos(tw.tPoint.pos());
        tw.tPoint.setLastScreenPos(tw.tPoint.screenPos());
        tw.tPoint.setLastScenePos(tw.tPoint.scenePos());
    }
    touchPointList.insert(newTouchPoint->id()+500, tw);
}

QList<QTouchEvent::TouchPoint> GhostView::getTouchpointList(){
    return listPtsOnGhostView.values();
}
